/* Filename: is_c.c
   Version: ANSI C Version
   Date:    November 1, 2019

   Author:  Martin Moehle
            Mathematical Institute
            University of Tuebingen
            Germany
            martin.moehle@uni-tuebingen.de

   Please compile this file using the command "gcc is_c.c -pedantic -Wall -O -o is_c.exe".

   This C program approximates the p-values of the homozygosity test and the
   Ewens-Watterson-Slatkin test using the importance sampling method
   described in 

   Moehle, M.: Simulation algorithms for integrals of a class of sampling distributions
   arising in population genetics, J. Stat. Comp. Simul. 75, 731-749. MR2145177

   For an analog program based on a Markov chain Monte Carlo method see mcmc_c.c. */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "data.h"     /* the data set c_obs */
#include "rand.h"     /* functions for random number generation */
#include "memerr.h"   /* function for memory problems */
#include "vecalloc.h" /* function which allocates memory for a configuration c */
#include "printcon.h" /* function for printing a configuration c */

/* constants for the function independentsampling */
#define M_MAX 100000  /* total number of simulations */
#define M_DOT 10000   /* after each M_DOT simulations a ot is printed */
#define M_DELTA 1e-10 /* shold be very small, but larger than the
			 machine exactness */

/* constants for the 95% confidence interval */
#define QUANTIL 1.96  /* 0.975 quantil of the standard normal distribution */

/* char *version="IS_C Version 1.0 (c) Martin Moehle 2002\n\n"; */

/* global variables of the program */
int n;         /* sample size */
int k;         /* number of alleles */
int *c;        /* configuration */
float h_obs;   /* the homozygosity of the observed configuration */
float p_obs;   /* the probability of the observed configuration */
float p_h,p_e; /* the values p_h and p_e for the homozygosity
		  test and for the EWS test */
float s_h,s_e; /* standard deviation of p_h and p_e */
float sigma2_h,/* p_h*(1-p_h) and */
      sigma2_e;/* p_e*(1-p_e) */
float alpha;   /* parameter for the logarithmic distribution */

float patil(int m, int n)     /* patil numbers */
{
   int i, j;
   float h;
   float *r; /* r is a pointer to an array of float variables */

   /* begin patil */
   if (n<0) return 0;
   if (m<n) return 0;
   /* 0 <= n <= m */
   if (m==n) return 1;
   if (n==0) return 0;
   /* 0 < n < m */

   /* allocate memory for buffer array r */
   if ((r=(float*)malloc((n+1)*sizeof(float)))==NULL) memerr("patil");
   r[0]=0; r[1]=1;
   for (i=2;i<=m;i++)
   {
      if (i<=n) r[i]=1;
      /* for j=min(n,i-1) downto max(i-m+n,1) */
      for (j=(n<i-1)?n:i-1;j>=i-m+n&&j>=1;j--) r[j]=(j*r[j-1]+(i-1)*r[j])/i;
   }
   h=r[n]; free(r); return h;
} /* end patil */

unsigned int logarithmic_rand(float a) {
/* algorithm LB in: A.W.Kemp (1981) Efficient generation of logarithmically
distributed pseudo-random variables, Appl. Stat. 30, 249-253.

The logarithmic distribution with parameter a can be viewed as a
geometric distribution G(P) with random parameter P, where
Q:=1-P is a continuous random variable with values in (0,a) and with
distribution function F(q) := ln(1-q) / ln(1-a), 0<q<a. Q can be
simulated easily via Q = 1-exp(U*ln(1-a)), where U is uniformly
distributed on the open unit interval (0,1). */
   return ceil(log(floatrand())/log(1.0-exp(floatrand()*log(1.0-a))));
} /* end logarithmic_rand */

unsigned int logarithmic_rand2(float a) {
/* algorithm LS in: A.W.Kemp (1981) Efficient generation of logarithmically
distributed pseudo-random variables, Appl. Stat. 30, 249-253 */
   unsigned int x;
   float p, u, t;
   t=-a/log(1.0-a);
   u=floatrand(); x=1; p=t;
   while (u>p) { u-=p; x+=1; p*=a*(x-1.0)/x; }
   return x;
} /* logarithmic_rand2 */

#define X_MAX 32767

unsigned int logarithmic_rand3(float a) {
/* inverse transform method with safety check if value gets too large */
   float b, c, k, u;
   unsigned int x;
   x=X_MAX;
   while (x >= X_MAX) {
      k=-1.0/log(1.0-a); c=k*a;
      b=c; u=floatrand(); x=1;
      while ((u>b) && (x<X_MAX)) {
	 c*=a; x+=1; b+=c/x;
	 if (x==X_MAX) {
	    printf("\nWarning: Function logarithmic_rand simulated a\n");
	    printf("         value equal to X_MAX = %d.\n",X_MAX);
	    printf("         The problem was solved by ignoring this\n");
	    printf("         value and simulating again.\n"); }
      } /* while */
   } /* while */
   return x;
} /* end logarithmic_rand3 */

float homozygosity(int *c, int k) {
   int i;
   float x,h;
   h=0;
   for (i=1;i<=k;i++) { x=(float)c[i]/n; h+=x*x; }
   return h;
} /* end homozygosity */

int allele(int n,int k) {
/* returns 1 (TRUE), if c satisfies c[1]+c[2]+...+c[k]=n,
otherwise 0 (FALSE) */
   int i,sum;
   sum=0; for(i=1;i<=k;i++) sum+=c[i];
   if (sum==n) return 1; else return 0;
} /* end allele */

float allele_dist(int n, int k, int *c) {
   int i;
   float h;
   h=1; for(i=1;i<=k;i++) h/=c[i];
   return h/patil(n,k);
} /* end allele_dist */

void initialize(void) {
   int i;
   /* initialize k and n */
   k=0;n=0;
   while (c_obs[k+1]) { k++; n+=c_obs[k]; }
   /* initialize c */
   c=vecalloc(k); for (i=1;i<=k;i++) c[i]=c_obs[i];
   h_obs=homozygosity(c,k);
   p_obs=allele_dist(n,k,c);
} /* end initialize */

int cond_rand(int j,int s) {
   double b,u;
   int x;
   x=1;
   b=patil(n-s-x,k-j) / (patil(n-s,k-j+1)*(float)x);
   u=floatrand();
   while (u>b) { x+=1; b+=patil(n-s-x,k-j)/(patil(n-s,k-j+1)*(float)x); }
   return x;
} /* end cond_rand */

void allele_rand() {
   int sum,j;
   sum=0;
   for(j=1;j<=k;j++) { c[j]=cond_rand(j,sum); sum+=c[j]; }
} /* end allele_rand */

float f(float x) {
   return (1-1/x)*log(1-x);
} /* end f */

float g(float y) {
   /* inverse of f, i.e. find 0<x<1 such that f(x)=y */
   float a,b,x;
   a=0;b=1;
   while (b-a>1e-5) {
      x=(a+b)/2;
      if (f(x)<y) b=x; else a=x;
   }
   return x;
} /* end g */

void allele_rand2(float alpha) {
   int sum;
   int i;
   /* the cases k=1 and k=n are trivial */
   if (k==1) c[1]=n; else
   if (k==n) for(i=1;i<=n;i++) c[i]=1; else {  /* 1<k<n */
      sum=0;
      while (sum!=n)
      {
	 sum=0; i=0;
	 while ((sum<=n) && (i<k)) {
	    i+=1;
	    c[i]=logarithmic_rand2(alpha);
/*	    printf(" %d " ,c[i]); getchar(); */
	    sum+=c[i];
	 } /* end while */
      } /* end while */
   } /* end else */
} /* end allele_rand2 */

void independentsampling(int *c, int k,
			 float *p_h, float *p_e,
			 float *s_h, float *s_e,
			 float *s2_h, float *s2_e) {
   unsigned long m;  /* index for counting all simulated configurations */
   /* variables for the homozygosity test */
   float h;           /* homozygosity of c */
   unsigned long m_h; /* number of simulated configurations with homozygosity */
		      /* smaller than or equal to the observed homozygosity */
   /* variables for the Ewens Watterson Slatkin test */
   float p;           /* probability of c */
   unsigned long m_e; /* number of simulated configurations with a smaller */
		      /* or equal probability than the observed configuration */
   m_h=0;
   m_e=0;
   fprintf(stderr,"   Please wait");
   for(m=1;m<=M_MAX;m++) {
      if (m%M_DOT==0) fprintf(stderr,"."); /* print a dot after each M_DOT simulations */
      allele_rand2(alpha);  /* use allele_rand(); or allele_rand2(alpha) */
/*      printconfig(c,k); */
      h=homozygosity(c,k);
      if (h<=h_obs+M_DELTA) m_h+=1;
      p=allele_dist(n,k,c);
      if (p<=p_obs+M_DELTA) m_e+=1;
   } /* end for m loop */
   *p_h=(float)m_h/M_MAX;   /* approximation for p_h */
   *p_e=(float)m_e/M_MAX;   /* approximation for p_e */
   *s2_h=*p_h*(1-*p_h);     /* approximation for the constant sigma^2 = mu*(1-mu) */
   *s2_e=*p_e*(1-*p_e);     /* for the homozygosity test and the EWS test */
   *s_h=sqrt(*s2_h/M_MAX);  /* approximation for the standard */
   *s_e=sqrt(*s2_e/M_MAX);  /* deviation of p_h and p_e */
   fprintf(stderr,"\n");
} /* end independentsampling */

int main() {
/* clrscr(); */ /* not in UNIX available */
   initialize(); /* define k,n,c,h_obs and p_obs */
   printf("sample size                 n = %d\n",n);
   printf("number of alleles           k = %d\n",k);
   printf("observed configuration  c_obs = "); printconfig(c,k);
   printf("homozygosity of c_obs   h_obs = %f\n",h_obs);
   printf("probability of c_obs    p_obs = %f\n",p_obs);
   printf("\nIndependent sampling method:\n\n");
   printf("   Number of simulations: %.0f\n",(float)M_MAX);
   printf("   A dot is printed after each %.0f simulations.\n",(float)M_DOT);
   printf("   M_DELTA = %e\n",(float)M_DELTA);
   /* the following parameter alpha is only needed, if the function
   "allele_rand2" is used in the function "independentsampling"; */
   /* alpha is chosen such that n=k/f(alpha) (=E(N_1+...+N_k)) */
   alpha=0; if ((k>1) && (k<n)) alpha=g((float)k/n);
   printf("   Parameter of the logarithmic distribution  alpha = %f\n",alpha);
   intrandomize(); floatrandomize(); /* initialize random number generators */
   independentsampling(c,k,&p_h,&p_e,&s_h,&s_e,&sigma2_h,&sigma2_e);
   printf("   Homozygosity test:            p_h    = %f\n",p_h);
   printf("   Ewens-Watterson-Slatkin test: p_e    = %f\n",p_e);
   printf("   Standard deviation for p_h           = %f\n",s_h);
   printf("   Standard deviation for p_e           = %f\n",s_e);
   printf("   95%% confidence interval for p_h      = [%f,%f]\n",p_h-s_h*QUANTIL,p_h+s_h*QUANTIL);
   printf("   95%% confidence interval for p_e      = [%f,%f]\n",p_e-s_e*QUANTIL,p_e+s_e*QUANTIL);
   printf("   Estimation for the constant sigma2_h = %f\n",sigma2_h);
   printf("   Estimation for the constant sigma2_e = %f\n",sigma2_e);
   free(c); /* dispose memory */
/*   printf("Press any key...");
   getch(); */  /* not in UNIX available */
   printf("Press <return> for exit...");
   getchar();
   return 1; /* most operating systems expect an int return value */
} /* end main */
