/* Filename: mcmc_c.c
   Version: ANSI C Version
   Date:    November 1, 2019

   Author:  Martin Moehle
            Mathematical Institute
            University of Tuebingen
            Germany
            martin.moehle@uni-tuebingen.de

   Please compile this file using the command "gcc mcmc_c.c -pedantic -Wall -O -o mcmc_c.exe".

   This C program approximates the p-values of the homozygosity test and the
   Ewens-Watterson-Slatkin test using the Markov chain Monte Carlo (MCMC) method
   described in 

   Moehle, M.: Simulation algorithms for integrals of a class of sampling distributions
   arising in population genetics, J. Stat. Comp. Simul. 75, 731-749. MR2145177

   For an analog program using importance sampling see is_c.c. */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#include "data.h"     /* the data set c_obs */
#include "rand.h"     /* functions for random number generation */
#include "memerr.h"   /* function for memory problems */
#include "vecalloc.h" /* function which allocates memory for a configuration c */
#include "printcon.h" /* function for printing a configuration c */

/* constant A >= 0 of the distribution pi_A */
#define A 0

/* constants for the function montecarlo */
#define M_MAX 10000000 /* total number of simulations */
#define M_MIN 50000    /* the first M_MIN simulations are not used for calculations */
#define M_DOT 100000   /* after each M_DOT simulations a dot is printed */
#define M_DELTA 1e-10  /* should be very small, but larger than the machine exactness */

/* constants for the 95% confidence interval */
#define QUANTIL 1.96   /* 0.975 quantil of the standard normal distribution */

/* global variables of the program */
int n;          /* sample size */
int k;          /* number of alleles */
int *c;         /* configuration */
float p_h,p_e;  /* the values p_h and p_e for the homozygosity
		   test and for the EWS test */
float s_h,s_e;  /* standard deviation of p_h and p_e */
float sigma2_h, /* the constants sigma^2 (see Corollary A.1 in the paper) */
      sigma2_e;
unsigned long
     jumps,     /* counter for the number of jumps of the Markov chain */
     visits;    /* counter for the number of visits of c_obs */

void initialize(void) {
   int i;
   /* initialize k and n */
   k=0;n=0;
   while (c_obs[k+1]) { k++; n+=c_obs[k]; }
   /* initialize c */
   c=vecalloc(k); for (i=1;i<=k;i++) c[i]=c_obs[i];
} /* end initialize */

int testequal(void) {
/* tests if c[1],...c[k] is equal to c_obs[1],...,c_obs[k] */
   int i;
   i=1;
   while ((c[i]==c_obs[i]) && (i<=k)) { i++; }
   if (i>k) return 1;
   return 0;
} /* end testequal */

/* a procedure to approximate p_h and p_e simultaneously */
void montecarlo(int *c, int k, float *p_h, float *p_e, float *s_h, float *s_e,
float *s2_h, float *s2_e, unsigned long *jumps, unsigned long *visits) {
/* input parameter:   c = the configuration stored in c[1],...,c[k]
		      k = number of alleles
   output parameters: p_h, p_e   = the p-value for the homozygosity test
				   and for the Ewens-Watterson-Slatkin test
		      jumps      = the number of jumps of the Markov chain
		      visits     = # of k, 1<=k<=M_MAX with X_k=c_obs
				 = the number of visits of the state c_obs
		      s_h, s_e   = standard deviation of p_h and p_e
		      s2_h, s2_e = the constant sigma^2 (see Corollary A.1
				   in the paper) for the Homozygosity test
				   and the EWS test */
   double q;          /* variable to save P(c(m))/P(c(m-1)) */
   int i,j;           /* indices for the two randomly chosen */
		      /* elements of the configuration */
   unsigned long m;   /* index for counting all simulated configurations */
   int equal;

   /* variables for the homozygosity test */
   long d;            /* variable for the difference (H(c(m))-H(c(0)))*n*n/2 */
   unsigned long m_h; /* number of simulated configurations with homozygosity */
		      /* smaller than or equal to the observed homozygosity */
   float z_h,e_h;     /* variables to calculate the standard deviation of p_h */

   /* variables for the Ewens Watterson Slatkin test */
   double rho;        /* varable for q(c(m))/q(c(0)) */
   unsigned long m_e; /* number of simulated configurations with a smaller */
		      /* or equal probability than the observed configuration */
   float z_e,e_e;     /* variables to calculate the standard deviation of p_e */

   rho=1.0;d=0;
   m_h=0;z_h=0;e_h=0;
   m_e=0;z_e=0;e_e=0;
   *jumps=0;*visits=0;
   fprintf(stderr,"   Please wait");
   for (m=1;m<=M_MAX;m++) {
      if (m%M_DOT==0) fprintf(stderr,"."); /* print a dot after each M_DOT simulations */
      i=intrand(k); j=intrand(k);          /* choose randomly distinct entries */
      while (i==j) j=intrand(k);           /* i and j with 1 <= i,j <= k */
      if (c[j]>1) {
	 q=(c[i]+A)/(c[i]+1.0)*c[j]/(c[j]+A-1.0);
	 if (floatrand()<q) {         /* with probability min(1,q) */
	    *jumps+=1;                /* count how often the chain jumps */
	    rho*=q;                   /* calculate rho=q(c(m+1))/q(c(0)) */
	    /* security check */
/*
	    if ((rho>1.0) && (rho<1.0+M_DELTA)) {
	       printf("\n   Warning: Possibly problems with M_DELTA. ");
	       printf("d = %d ",2*d); printf(" rho=%f=%e\n",rho,rho);
	    }
*/
	    c[i]+=1;
	    d+=c[i]-c[j];          /* calculate d=(H(c(m+1))-H(c(0)))*n*n/2 */
	    c[j]-=1; } }           /* and move to the next configuration. */
      equal=testequal();
      if (equal) { *visits+=1; }   /* count the number of visits of c_obs */

      if (m>M_MIN) {
	 if (d<=0) {
	    m_h+=1; z_h+=1;        /* increment counters for the homozygosity */
	 }
	 if (rho<=1.0+M_DELTA) {
	    m_e+=1; z_e+=1;        /* test and for the EWS test */
	 }
	 z_h-=(float)m_h/(m-M_MIN);/* the mean has to be substracted */
	 z_e-=(float)m_e/(m-M_MIN);
	 if (equal) {
	    e_h+=z_h*z_h;          /* update e = sum z^2 for the homozygosity */
	    e_e+=z_e*z_e;          /* test and for the EWS test and */
	    z_h=0;z_e=0;           /* reset z_h and z_e */
	 }
      } /* if m>M_MIN */
   } /* end for m loop */
   *p_h=((float) m_h/(M_MAX-M_MIN));  /* approximation for p_h */
   *p_e=((float) m_e/(M_MAX-M_MIN));  /* approximation for p_e */
   *s_h=sqrt(e_h)/(M_MAX-M_MIN);      /* approximation for the standard */
   *s_e=sqrt(e_e)/(M_MAX-M_MIN);      /* deviation of p_h and p_e */
   *s2_h=e_h/(M_MAX-M_MIN);           /* estimation for the constant sigma^2 */
   *s2_e=e_e/(M_MAX-M_MIN);           /* for the homozygosity test and the EWS test */
   fprintf(stderr,"\n");
} /* end montecarlo */

int main(void) {
   initialize(); /* define k,n and c */
   printf("Distribution pi_A with      A = %f\n",(float)A);
   printf("Sample size                 n = %d\n",n);
   printf("Number of alleles           k = %d\n",k);
   printf("Observed configuration  c_obs = "); printconfig(c,k);
   printf("\nMarkov Chain Monte Carlo method:\n\n");
   printf("   Number of simulations: %.0f\n",(float)M_MAX);
   printf("   The first %.0f simulations are not used for calculations.\n",(float)M_MIN);
   printf("   A dot is printed after each %.0f simulations.\n",(float)M_DOT);
   printf("   M_DELTA = %e\n",(float)M_DELTA);
   intrandomize(); floatrandomize(); /* initialize random number generators */
   montecarlo(c,k,&p_h,&p_e,&s_h,&s_e,&sigma2_h,&sigma2_e,&jumps,&visits);
   printf("   Homozygosity test:            p_h    = %f\n",p_h);
   printf("   Ewens-Watterson-Slatkin test: p_e    = %f\n",p_e);
   printf("   Standard deviation for p_h           = %f\n",s_h);
   printf("   Standard deviation for p_e           = %f\n",s_e);
   printf("   95%% confidence interval for p_h      = [%f,%f]\n",p_h-s_h*QUANTIL,p_h+s_h*QUANTIL);
   printf("   95%% confidence interval for p_e      = [%f,%f]\n",p_e-s_e*QUANTIL,p_e+s_e*QUANTIL);
   printf("   Estimation for the constant sigma2_h = %f\n",sigma2_h);
   printf("   Estimation for the constant sigma2_e = %f\n",sigma2_e);
   printf("   The Markov chain jumped %.0f times and\n",(float)jumps);
   printf("   visited state c_obs %.0f times.",(float)visits);
   free(c); /* dispose memory */
   return 1; /* most operating systems expect an int return value */
} /* end main */
